# mypy: disable-error-code="no-untyped-def"
import json
import urllib.parse
from typing import cast

import pytest
import requests
from lxml import etree
from rdflib import RDF
from rdflib import XSD
from rdflib import Graph
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.utils.rdflib_constants import KNORA_API
from dsp_tools.utils.rdflib_constants import KNORA_API_STR
from dsp_tools.utils.rdflib_constants import SubjectObjectTypeAlias
from test.e2e.commands.xmlupload.utils import util_get_res_iri_from_label
from test.e2e.commands.xmlupload.utils import util_request_resources_by_class

BASE_NUMBER_OF_TRIPLES_PER_VALUE = 9
OPEN_PERMISSIONS = Literal(
    "CR knora-admin:ProjectAdmin|D knora-admin:ProjectMember|V knora-admin:KnownUser,knora-admin:UnknownUser"
)
DOAP_PERMISSIONS = Literal("CR knora-admin:ProjectAdmin|D knora-admin:ProjectMember")

RICHTEXT_XML_DECLARATION = '<?xml version="1.0" encoding="UTF-8"?>\n'

SPECIAL_CHARACTERS_STRING = "àéèêëôûç äöüß _-'()[]{}+=!| 漢が글ርبيةб中זרקצחק §µÞðΘΨ∉∴∫⊗‰♦"


@pytest.fixture(scope="module")
def g_minimal(_xmlupload_minimal_correct, class_with_everything_iri, auth_header, project_iri, creds) -> Graph:
    return util_request_resources_by_class(class_with_everything_iri, auth_header, project_iri, creds)


@pytest.fixture(scope="module")
def g_text_parsing(_xmlupload_text_parsing, class_with_everything_iri, auth_header, project_iri, creds) -> Graph:
    return util_request_resources_by_class(class_with_everything_iri, auth_header, project_iri, creds)


def _assert_number_of_values_is_one_and_get_val_iri(g: Graph, label: str, prop_iri: URIRef) -> SubjectObjectTypeAlias:
    res_iri = util_get_res_iri_from_label(g, label)
    val_iri_list = list(g.objects(res_iri, prop_iri))
    assert len(val_iri_list) == 1
    return val_iri_list.pop(0)


class TestSharedTriples:
    """
    The xmlupload code that creates metadata triples (e.g. the permissions) is shared by all value types.
    These generic triples are tested only once and not for each value type individually.
    """

    def test_all_triples_of_a_value(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testBoolean")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "bool_true", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        props = [
            KNORA_API.arkUrl,
            KNORA_API.versionArkUrl,
            KNORA_API.userHasPermission,
            KNORA_API.attachedToUser,
            KNORA_API.valueHasUUID,
            KNORA_API.hasPermissions,
            RDF.type,
            KNORA_API.booleanValueAsBoolean,
            KNORA_API.valueCreationDate,
        ]
        for prp in props:
            assert len(list(g_minimal.objects(val_iri, prp))) == 1
        assert len(val_triples) == len(props)

    def test_doap(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testBoolean")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "bool_true", prop_iri)
        permissions = next(g_minimal.objects(val_iri, KNORA_API.hasPermissions))
        assert permissions == DOAP_PERMISSIONS

    def test_open_permissions(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testBoolean")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(
            g_minimal, "bool_1_open_permissions_on_value", prop_iri
        )
        permissions = next(g_minimal.objects(val_iri, KNORA_API.hasPermissions))
        assert permissions == OPEN_PERMISSIONS

    def test_comment_on_value(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testBoolean")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "bool_0_comment_on_value", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        cmnt = next(g_minimal.objects(val_iri, KNORA_API.valueHasComment))
        assert cmnt == Literal("Comment on value")
        comment_triple = 1
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE + comment_triple


class TestValues:
    """
    The following tests, test the triples that are specific to each value type.
    """

    def test_bool_true(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testBoolean")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "bool_true", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        expected_val = Literal(True)
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.booleanValueAsBoolean))
        assert actual_value == expected_val
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.BooleanValue
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE

    def test_color(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testColor")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "color", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        expected_val = Literal("#00ff00")
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.colorValueAsColor))
        assert actual_value == expected_val
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.ColorValue
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE

    def test_date(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testSubDate1")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "date", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        expected_val = Literal("JULIAN:0700 BCE:0600 BCE")
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.valueAsString))
        assert actual_value == expected_val
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.DateValue
        date_details = [
            (KNORA_API.dateValueHasCalendar, Literal("JULIAN")),
            (KNORA_API.dateValueHasStartEra, Literal("BCE")),
            (KNORA_API.dateValueHasStartYear, Literal("700", datatype=XSD.integer)),
            (KNORA_API.dateValueHasEndEra, Literal("BCE")),
            (KNORA_API.dateValueHasEndYear, Literal("600", datatype=XSD.integer)),
        ]
        for prp, val in date_details:
            assert next(g_minimal.objects(val_iri, prp)) == val
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE + len(date_details)

    def test_decimal(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testDecimalSimpleText")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "decimal", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        expected_val = Literal("2.71", datatype=XSD.decimal)
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.decimalValueAsDecimal))
        assert actual_value == expected_val
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.DecimalValue
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE

    def test_geometry(self, auth_header, project_iri, creds):
        cls_iri_str = f"{KNORA_API_STR}Region"
        res_g = util_request_resources_by_class(cls_iri_str, auth_header, project_iri, creds)
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(res_g, "region", KNORA_API.hasGeometry)
        val_triples = list(res_g.triples((val_iri, None, None)))
        actual_value = str(next(res_g.objects(val_iri, KNORA_API.geometryValueAsGeometry)))
        geom_json = json.loads(actual_value)
        assert set(geom_json.keys()) == {"status", "type", "lineWidth", "points"}
        assert next(res_g.objects(val_iri, RDF.type)) == KNORA_API.GeomValue
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE

    def test_geoname(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testGeoname")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "geoname", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        expected_val = Literal("1111111")
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.geonameValueAsGeonameCode))
        assert actual_value == expected_val
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.GeonameValue
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE

    def test_integer(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testIntegerSimpleText")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "integer", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        expected_val = Literal(1)
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.intValueAsInt))
        assert actual_value == expected_val
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.IntValue
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE

    def test_interval(self, auth_header, project_iri, creds):
        cls_iri_str = f"{KNORA_API_STR}AudioSegment"
        res_g = util_request_resources_by_class(cls_iri_str, auth_header, project_iri, creds)
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(res_g, "audio_segment", KNORA_API.hasSegmentBounds)
        val_triples = list(res_g.triples((val_iri, None, None)))
        interval_start = Literal("0.1", datatype=XSD.decimal)
        actual_interval_start = next(res_g.objects(val_iri, KNORA_API.intervalValueHasStart))
        assert actual_interval_start == interval_start
        interval_end = Literal("0.2", datatype=XSD.decimal)
        actual_interval_end = next(res_g.objects(val_iri, KNORA_API.intervalValueHasEnd))
        assert actual_interval_end == interval_end
        assert next(res_g.objects(val_iri, RDF.type)) == KNORA_API.IntervalValue
        additional_interval_triple = 1
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE + additional_interval_triple

    def _util_get_list_node(self, creds, auth_header) -> str:
        all_lists_one_project_endpoint = f"{creds.server}/admin/lists?9999"
        headers = auth_header | {"Content-Type": "application/json"}
        response_all_lists = requests.get(all_lists_one_project_endpoint, timeout=3, headers=headers).json()
        first_list = next(x for x in response_all_lists["lists"] if x["name"] == "firstList")
        list_iri = urllib.parse.quote_plus(first_list["id"])
        all_nodes_one_list_endpoint = f"{creds.server}/admin/lists/{list_iri}"
        response_all_nodes = requests.get(all_nodes_one_list_endpoint, timeout=3, headers=headers).json()
        children = response_all_nodes["list"]["children"]
        node_one = list(x for x in children if x["name"] == "n1")
        assert len(node_one) == 1
        return cast(str, node_one.pop(0)["id"])

    def test_list(self, g_minimal, onto_iri, creds, auth_header):
        prop_iri = URIRef(f"{onto_iri}testListProp")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "list", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        list_node_iri = self._util_get_list_node(creds, auth_header)
        expected_val = URIRef(list_node_iri)
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.listValueAsListNode))
        assert actual_value == expected_val
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.ListValue
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE

    def test_link(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testHasLinkToValue")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "link", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        target_iri = util_get_res_iri_from_label(g_minimal, "resource_no_values")
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.linkValueHasTargetIri))
        assert actual_value == target_iri
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.LinkValue
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE

    def test_richtext(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testRichtext")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "richtext", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        expected_val = Literal(f"{RICHTEXT_XML_DECLARATION}<text><p> Text </p></text>")
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.textValueAsXml))
        assert actual_value == expected_val
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.TextValue
        text_type = next(g_minimal.objects(val_iri, KNORA_API.hasTextValueType))
        assert text_type == KNORA_API.FormattedText
        mapping_type = next(g_minimal.objects(val_iri, KNORA_API.textValueHasMapping))
        assert mapping_type == URIRef("http://rdfh.ch/standoff/mappings/StandardMapping")
        additional_richtext_triples = 2
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE + additional_richtext_triples

    def test_textarea(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testTextarea")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "textarea", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        expected_val = Literal("Line One\nLine Two")
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.valueAsString))
        assert actual_value == expected_val
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.TextValue
        text_type = next(g_minimal.objects(val_iri, KNORA_API.hasTextValueType))
        assert text_type == KNORA_API.UnformattedText
        additional_text_type_triple = 1
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE + additional_text_type_triple

    def test_simpletext(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testSimpleText")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "simpletext", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        expected_val = Literal("Text")
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.valueAsString))
        assert actual_value == expected_val
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.TextValue
        text_type = next(g_minimal.objects(val_iri, KNORA_API.hasTextValueType))
        assert text_type == KNORA_API.UnformattedText
        additional_text_type_triple = 1
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE + additional_text_type_triple

    def test_time(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testTimeValue")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "time", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        expected_val = Literal("2019-10-24T03:45:12.010Z", datatype=XSD.dateTimeStamp)
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.timeValueAsTimeStamp))
        assert actual_value == expected_val
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.TimeValue
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE

    def test_uri(self, g_minimal, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testUriValue")
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g_minimal, "uri", prop_iri)
        val_triples = list(g_minimal.triples((val_iri, None, None)))
        expected_val = Literal("https://dasch.swiss", datatype=XSD.anyURI)
        actual_value = next(g_minimal.objects(val_iri, KNORA_API.uriValueAsUri))
        assert actual_value == expected_val
        assert next(g_minimal.objects(val_iri, RDF.type)) == KNORA_API.UriValue
        assert len(val_triples) == BASE_NUMBER_OF_TRIPLES_PER_VALUE


class TestTextParsing:
    """
    This tests if the content of text is correctly parsed (tags, special characters, etc.).
    It does not test if the value is well-formed as that is done separately.
    All richtext values start with an XML header and are wrapped in a <text> tag which is not in the original data.
    """

    def _util_get_string_value(
        self, g: Graph, res_label: str, prop_iri: URIRef, knora_api_prop: URIRef = KNORA_API.textValueAsXml
    ) -> str:
        val_iri = _assert_number_of_values_is_one_and_get_val_iri(g, res_label, prop_iri)
        actual_value = list(g.objects(val_iri, knora_api_prop))
        assert len(actual_value) == 1
        return str(actual_value.pop(0))

    def test_richtext_res_with_tags_in_text(self, g_text_parsing, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testRichtext")
        returned_str = self._util_get_string_value(g_text_parsing, "res_with_tags_in_text", prop_iri)
        expected_str = (
            f"{RICHTEXT_XML_DECLARATION}<text>This is <em>italicized and <strong>bold</strong> </em> text!</text>"
        )
        assert returned_str == expected_str

    def test_richtext_res_text_wrapped_in_tag(self, g_text_parsing, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testRichtext")
        returned_str = self._util_get_string_value(g_text_parsing, "res_text_wrapped_in_tag", prop_iri)
        expected_str = f"{RICHTEXT_XML_DECLARATION}<text><p> Paragraph text </p></text>"
        assert returned_str == expected_str

    def test_richtext_res_with_multiple_paragraphs(self, g_text_parsing, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testRichtext")
        returned_str = self._util_get_string_value(g_text_parsing, "res_with_multiple_paragraphs", prop_iri)
        expected_str = f"{RICHTEXT_XML_DECLARATION}<text><p> Paragraph 1 text </p> <p> Paragraph 2 text </p></text>"
        assert returned_str == expected_str

    def test_richtext_res_with_escaped_characters(self, g_text_parsing, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testRichtext")
        returned_str = self._util_get_string_value(g_text_parsing, "res_with_escaped_characters", prop_iri)
        expected_str = f"{RICHTEXT_XML_DECLARATION}<text>text &lt;notatag&gt; text and with ampersand &amp;</text>"
        assert returned_str == expected_str

    def test_richtext_res_with_standoff_link_to_id(self, g_text_parsing, onto_iri):
        # In the order of the attributes in the <a> tag to denote a stand-off is not always the same.
        # Therefore, a string comparison may fail at some times.
        target_iri = util_get_res_iri_from_label(g_text_parsing, "target_resource_with_id")
        prop_iri = URIRef(f"{onto_iri}testRichtext")
        returned_str = self._util_get_string_value(g_text_parsing, "richtext_res_with_standoff_link_to_id", prop_iri)
        returned_str = returned_str.removeprefix(RICHTEXT_XML_DECLARATION)
        returned_tree = etree.fromstring(returned_str)
        assert not returned_tree.text
        emphasis = next(returned_tree.iter(tag="em"))
        assert emphasis.text == "Text "
        assert emphasis.tail == " and some tags"
        stand_off_link = next(emphasis.iter(tag="a"))
        assert stand_off_link.text == "target_resource_with_id"
        assert stand_off_link.attrib["class"] == "salsah-link"
        assert stand_off_link.attrib["href"] == str(target_iri)

    def test_richtext_res_with_standoff_link_to_iri(self, g_text_parsing, onto_iri):
        # When serialised as string, the order of the attributes in the <a> tag is not always the same.
        # Therefore, we must test against the etree object, not against a string.
        prop_iri = URIRef(f"{onto_iri}testRichtext")
        returned_str = self._util_get_string_value(g_text_parsing, "res_with_standoff_link_to_iri", prop_iri)
        returned_str = returned_str.removeprefix(RICHTEXT_XML_DECLARATION)
        returned_tree = etree.fromstring(returned_str)
        assert returned_tree.text == "Text "
        stand_off_link = next(returned_tree.iter(tag="a"))
        assert stand_off_link.text == "target_resource_with_iri"
        assert stand_off_link.attrib["class"] == "salsah-link"
        assert stand_off_link.attrib["href"] == "http://rdfh.ch/9999/DiAmYQzQSzC7cdTo6OJMYA"
        assert stand_off_link.tail == " end text"

    def test_richtext_res_with_standoff_link_to_url(self, g_text_parsing, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testRichtext")
        returned_str = self._util_get_string_value(g_text_parsing, "richtext_res_with_standoff_link_to_url", prop_iri)
        expected_str = (
            f'{RICHTEXT_XML_DECLARATION}<text>Text <a href="https://www.dasch.swiss/">URL</a> end text</text>'
        )
        assert returned_str == expected_str

    def test_richtext_res_with_footnotes(self, g_text_parsing, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testRichtext")
        returned_str = self._util_get_string_value(g_text_parsing, "res_with_footnotes", prop_iri)
        expected_str = f'{RICHTEXT_XML_DECLARATION}<text>Text <footnote content="Footnote"/> end text</text>'
        assert returned_str == expected_str

    def test_richtext_res_with_escaped_chars_in_footnote(self, g_text_parsing, onto_iri):
        """
        Currently, DSP-API doesn't create this value correctly,
        see [DEV-4796](https://linear.app/dasch/issue/DEV-4796).
        This test succeeds currently, but as soon as the bug is fixed, it will fail.
        The correct API response would be `href=&quot;` instead of `href=\\&quot;`.
        This is also true for the single apostrophe in the following two tests.
        Use this ticket to fix the tests: https://linear.app/dasch/issue/DEV-4878/add-footnote-parsing-test-in-xmlupload
        """
        prop_iri = URIRef(f"{onto_iri}testRichtext")
        returned_str = self._util_get_string_value(g_text_parsing, "res_with_escaped_chars_in_footnote", prop_iri)
        expected_str = (
            f"{RICHTEXT_XML_DECLARATION}<text>"
            f'Text <footnote content="Text &lt;a href=\\&quot;https://www.google.com/\\&quot;&gt;Google&lt;/a&gt;"/> '
            f"end text</text>"
        )
        assert returned_str == expected_str

    def test_special_characters_in_richtext(self, g_text_parsing, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testRichtext")
        returned_str = self._util_get_string_value(g_text_parsing, "res_richtext_special_characters", prop_iri)
        wrongly_escaped_special_char_string = "àéèêëôûç äöüß _-&apos;()[]{}+=!| 漢が글ርبيةб中זרקצחק §µÞðΘΨ∉∴∫⊗‰♦"
        expected_str = f"{RICHTEXT_XML_DECLARATION}<text>{wrongly_escaped_special_char_string}</text>"
        assert returned_str == expected_str

    def test_special_characters_in_footnote(self, g_text_parsing, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testRichtext")
        returned_str = self._util_get_string_value(g_text_parsing, "res_special_chars_in_footnote", prop_iri)
        wrongly_escaped_special_char_string = "àéèêëôûç äöüß _-\&apos;()[]{}+=!| 漢が글ርبيةб中זרקצחק §µÞðΘΨ∉∴∫⊗‰♦"
        expected_str = (
            f'{RICHTEXT_XML_DECLARATION}<text>Text <footnote content="{wrongly_escaped_special_char_string}"/> '
            f"end text</text>"
        )
        assert returned_str == expected_str

    def test_special_characters_in_simpletext(self, g_text_parsing, onto_iri):
        prop_iri = URIRef(f"{onto_iri}testSimpleText")
        returned_str = self._util_get_string_value(
            g_text_parsing, "res_simpletext_special_characters", prop_iri, KNORA_API.valueAsString
        )
        assert returned_str == SPECIAL_CHARACTERS_STRING

    def test_special_characters_in_label(self, g_text_parsing):
        # Since we search the resource according to its label we know that it was parsed correctly.
        # Otherwise, this function call would fail.
        returned_iri = util_get_res_iri_from_label(g_text_parsing, SPECIAL_CHARACTERS_STRING)
        assert str(returned_iri).startswith("http://rdfh.ch/9999/")



from dsp_tools.clients.connection import Connection


class Model:
    """Base class for other classes"""

    _con: Connection
    _changed: set[str]

    def __init__(self, con: Connection):
        self._con = con
        self._changed = set()



from dataclasses import dataclass
from dataclasses import field

from dsp_tools.commands.xmlupload.iri_resolver import IriResolver
from dsp_tools.commands.xmlupload.models.processed.res import ProcessedResource
from dsp_tools.commands.xmlupload.stash.stash_models import Stash
from dsp_tools.commands.xmlupload.upload_config import UploadConfig


@dataclass
class UploadState:
    """
    Save the state of an xmlupload, so that after an interruption, it can be resumed.
    """

    pending_resources: list[ProcessedResource]
    pending_stash: Stash | None
    config: UploadConfig
    failed_uploads: list[str] = field(default_factory=list)
    iri_resolver: IriResolver = field(default_factory=IriResolver)



# mypy: disable-error-code="method-assign,no-untyped-def"

from rdflib import RDF
from rdflib import SH
from rdflib import Literal
from rdflib import URIRef

from dsp_tools.commands.validate_data.models.api_responses import EnabledLicenseIris
from dsp_tools.commands.validate_data.sparql.legal_info_shacl import construct_allowed_licenses_shape
from dsp_tools.utils.rdflib_constants import API_SHAPES
from dsp_tools.utils.rdflib_constants import KNORA_API

NUMBER_OF_TRIPLES_NO_LICENSES = 8


def test_construct_allowed_licenses_shape_one_license():
    all_licenses = EnabledLicenseIris(["http://rdfh.ch/1"])
    result = construct_allowed_licenses_shape(all_licenses)
    referenced_license_triples = 2
    assert len(result) == NUMBER_OF_TRIPLES_NO_LICENSES + referenced_license_triples
    subj = API_SHAPES.FileValue_ClassShape
    assert next(result.objects(subj, RDF.type)) == SH.NodeShape
    assert next(result.objects(subj, SH.targetClass)) == KNORA_API.FileValue
    prop_shape = next(result.objects(subj, SH.property))
    assert next(result.objects(prop_shape, RDF.type)) == SH.PropertyShape
    assert next(result.objects(prop_shape, SH.path)) == KNORA_API.hasLicense
    in_bn, license_iris = next(result.subject_objects(predicate=RDF.first))
    sh_in = URIRef("http://www.w3.org/ns/shacl#in")
    assert next(result.objects(prop_shape, sh_in)) == in_bn
    assert license_iris == URIRef("http://rdfh.ch/1")
    expected_msg = (
        "Please only use enabled licenses in your data. Consult the project information for enabled licenses."
    )
    assert next(result.objects(prop_shape, SH.message)) == Literal(expected_msg)
    assert next(result.objects(prop_shape, SH.severity)) == SH.Violation


def test_construct_allowed_licenses_shape_no_license():
    all_licenses = EnabledLicenseIris([])
    result = construct_allowed_licenses_shape(all_licenses)
    assert len(result) == NUMBER_OF_TRIPLES_NO_LICENSES
    subj = API_SHAPES.FileValue_ClassShape
    assert next(result.objects(subj, RDF.type)) == SH.NodeShape
    assert next(result.objects(subj, SH.targetClass)) == KNORA_API.FileValue
    prop_shape = next(result.objects(subj, SH.property))
    assert next(result.objects(prop_shape, RDF.type)) == SH.PropertyShape
    assert next(result.objects(prop_shape, SH.path)) == KNORA_API.hasLicense
    expected_msg = "You are only allowed to reference enabled licenses. No licenses are enabled for this project."
    assert next(result.objects(prop_shape, SH.message)) == Literal(expected_msg)
    assert next(result.objects(prop_shape, SH.severity)) == SH.Violation



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class ProjectDefinition:
    metadata: ProjectMetadata
    ontologies: list[dict[str, Any]]
    lists: list[dict[str, Any]] | None
    groups: list[dict[str, str]] | None
    users: list[dict[str, str]] | None


@dataclass
class ProjectMetadata:
    shortcode: str
    shortname: str
    longname: str
    keywords: list[str] | None = None
    descriptions: dict[str, str] | None = None
    enabled_licenses: list[str] | None = None



# mypy: disable-error-code="no-untyped-def,comparison-overlap"

from lxml import etree

from dsp_tools.xmllib.internal.constants import DASCH_SCHEMA
from dsp_tools.xmllib.models.config_options import Permissions
from dsp_tools.xmllib.models.dsp_base_resources import AudioSegmentResource
from dsp_tools.xmllib.models.dsp_base_resources import LinkResource
from dsp_tools.xmllib.models.dsp_base_resources import RegionResource
from dsp_tools.xmllib.models.dsp_base_resources import VideoSegmentResource
from dsp_tools.xmllib.models.licenses.recommended import LicenseRecommended
from dsp_tools.xmllib.models.res import Resource
from dsp_tools.xmllib.models.root import XMLRoot
from dsp_tools.xmllib.models.root import _make_authorship_lookup
from dsp_tools.xmllib.models.root import _serialise_authorship


class TestSerialise:
    def test_regular_resource(self):
        xml_root = XMLRoot.create_new("0000", "test")
        xml_root.add_resource(
            Resource.create_new(
                "res_id",
                ":Restype",
                "lbl",
            ).add_bool(":boolProp", True, comment="cmnt")
        )
        serialised = xml_root.serialise()
        found = list(serialised.iterdescendants(tag=f"{DASCH_SCHEMA}resource"))
        assert len(found) == 1
        resource = found[0]
        res_attribs = {"label": "lbl", "restype": ":Restype", "id": "res_id"}
        assert resource.attrib == res_attribs
        value_list = list(resource.iterchildren())
        assert len(value_list) == 1
        bool_prop = value_list[0]
        assert bool_prop.tag == f"{DASCH_SCHEMA}boolean-prop"
        assert bool_prop.attrib == {"name": ":boolProp"}
        bool_prop_values = list(bool_prop.iterchildren())
        assert len(bool_prop_values) == 1
        prop_val = bool_prop_values[0]
        assert prop_val.tag == f"{DASCH_SCHEMA}boolean"
        assert prop_val.text == "true"
        assert prop_val.attrib == {"comment": "cmnt"}

    def test_region(self):
        xml_root = XMLRoot.create_new("0000", "test")
        xml_root.add_resource(
            RegionResource.create_new(
                "region",
                "lbl",
                "region_of_val",
                Permissions.RESTRICTED,
            ).add_rectangle((0.1, 0.1), (0.2, 0.2))
        )
        serialised = xml_root.serialise()
        found = list(serialised.iterdescendants(tag=f"{DASCH_SCHEMA}region"))
        assert len(found) == 1
        resource = found[0]
        res_attribs = {"label": "lbl", "id": "region", "permissions": "restricted"}
        assert resource.attrib == res_attribs
        assert len(resource) == 3
        # Check color
        all_colors = list(resource.iterchildren(tag=f"{DASCH_SCHEMA}color-prop"))
        assert len(all_colors) == 1
        color_prop = all_colors[0]
        assert color_prop.attrib == {"name": "hasColor"}
        color_child = next(color_prop.iterchildren())
        assert color_child.text == "#5b24bf"
        assert color_child.attrib == {"permissions": "restricted"}
        # Check is region of
        all_resptrs = list(resource.iterchildren(tag=f"{DASCH_SCHEMA}resptr-prop"))
        assert len(all_resptrs) == 1
        resptr_prop = all_resptrs[0]
        assert resptr_prop.attrib == {"name": "isRegionOf"}
        resptr_child = next(resptr_prop.iterchildren())
        assert resptr_child.text == "region_of_val"
        assert resptr_child.attrib == {"permissions": "restricted"}
        # Check geometry
        all_geo = list(resource.iterchildren(tag=f"{DASCH_SCHEMA}geometry-prop"))
        assert len(all_geo) == 1
        geo_prop = all_geo[0]
        assert geo_prop.attrib == {"name": "hasGeometry"}
        geo_child = next(geo_prop.iterchildren())
        assert geo_child.text
        assert len(geo_child.text) > 0
        assert geo_child.attrib == {"permissions": "restricted"}

    def test_link_resource(self):
        xml_root = XMLRoot.create_new("0000", "test")
        xml_root.add_resource(
            LinkResource.create_new(
                "link_id",
                "lbl",
                ["link1", "link2"],
            ).add_comment("cmnt", Permissions.RESTRICTED)
        )
        serialised = xml_root.serialise()
        found = list(serialised.iterdescendants(tag=f"{DASCH_SCHEMA}link"))
        assert len(found) == 1
        resource = found[0]
        res_attribs = {"label": "lbl", "id": "link_id"}
        assert resource.attrib == res_attribs
        vals = list(resource.iterchildren())
        assert len(vals) == 2
        # link tos
        all_resptrs = list(resource.iterchildren(tag=f"{DASCH_SCHEMA}resptr-prop"))
        assert len(all_resptrs) == 1
        resptr = all_resptrs[0]
        assert resptr.attrib == {"name": "hasLinkTo"}
        for v in resptr.iterchildren():
            assert v.tag == f"{DASCH_SCHEMA}resptr"
            assert v.text in ["link1", "link2"]
            assert not v.attrib
        # comment
        all_text = list(resource.iterchildren(tag=f"{DASCH_SCHEMA}text-prop"))
        assert len(all_text) == 1
        text_prop = all_text[0]
        assert text_prop.attrib == {"name": "hasComment"}
        text_child = next(text_prop.iterchildren())
        assert text_child.text == "cmnt"
        assert text_child.attrib == {"permissions": "restricted", "encoding": "xml"}

    def test_audio_segment(self):
        xml_root = XMLRoot.create_new("0000", "test")
        xml_root.add_resource(
            AudioSegmentResource.create_new("audio_id", "lbl", "segment_of", 1, 2)
            .add_title("title")
            .add_comment("cmnt")
            .add_description("desc")
            .add_keyword("keywrd")
            .add_relates_to("relates_to")
        )
        serialised = xml_root.serialise()
        found = list(serialised.iterdescendants(tag=f"{DASCH_SCHEMA}audio-segment"))
        assert len(found) == 1
        resource = found[0]
        res_attribs = {"label": "lbl", "id": "audio_id"}
        assert resource.attrib == res_attribs
        assert len(resource) == 7
        segment_of = next(resource.iter(tag=f"{DASCH_SCHEMA}isSegmentOf"))
        assert not segment_of.attrib
        assert segment_of.text == "segment_of"
        bounds = next(resource.iter(tag=f"{DASCH_SCHEMA}hasSegmentBounds"))
        assert not bounds.text
        assert bounds.attrib == {"segment_start": "1", "segment_end": "2"}
        title = next(resource.iter(tag=f"{DASCH_SCHEMA}hasTitle"))
        assert not title.attrib
        assert title.text == "title"
        comment = next(resource.iter(tag=f"{DASCH_SCHEMA}hasComment"))
        assert not comment.attrib
        assert comment.text == "cmnt"
        description = next(resource.iter(tag=f"{DASCH_SCHEMA}hasDescription"))
        assert not description.attrib
        assert description.text == "desc"
        keywrd = next(resource.iter(tag=f"{DASCH_SCHEMA}hasKeyword"))
        assert not keywrd.attrib
        assert keywrd.text == "keywrd"
        relates_to = next(resource.iter(tag=f"{DASCH_SCHEMA}relatesTo"))
        assert not relates_to.attrib
        assert relates_to.text == "relates_to"

    def test_video_segment_with_permissions(self):
        xml_root = XMLRoot.create_new("0000", "test")
        xml_root.add_resource(
            VideoSegmentResource.create_new(
                "video_id",
                "lbl",
                "segment_of",
                1,
                2,
                permissions=Permissions.RESTRICTED,
            )
        )
        serialised = xml_root.serialise()
        found = list(serialised.iterdescendants(tag=f"{DASCH_SCHEMA}video-segment"))
        assert len(found) == 1
        resource = found[0]
        res_attribs = {"label": "lbl", "id": "video_id", "permissions": "restricted"}
        assert resource.attrib == res_attribs
        assert len(resource) == 2
        segment_of = next(resource.iter(tag=f"{DASCH_SCHEMA}isSegmentOf"))
        assert segment_of.attrib == {"permissions": "restricted"}
        assert segment_of.text == "segment_of"
        bounds = next(resource.iter(tag=f"{DASCH_SCHEMA}hasSegmentBounds"))
        assert not bounds.text
        assert bounds.attrib == {"segment_start": "1", "segment_end": "2", "permissions": "restricted"}


class TestSerialiseOverwriteDefaultPermissions:
    def test_regular_resource(self):
        xml_root = XMLRoot.create_new("0000", "test")
        xml_root.add_resource(
            Resource.create_new(
                "res_id",
                ":Restype",
                "lbl",
            ).add_bool(":boolProp", True, comment="cmnt")
        )
        serialised = xml_root.serialise(default_permissions=Permissions.OPEN)
        found = list(serialised.iterdescendants(tag=f"{DASCH_SCHEMA}resource"))
        assert len(found) == 1
        resource = found[0]
        res_attribs = {"label": "lbl", "restype": ":Restype", "id": "res_id", "permissions": "open"}
        assert resource.attrib == res_attribs
        value_list = list(resource.iterchildren())
        assert len(value_list) == 1
        bool_prop = value_list[0]
        assert bool_prop.tag == f"{DASCH_SCHEMA}boolean-prop"
        assert bool_prop.attrib == {"name": ":boolProp"}
        bool_prop_values = list(bool_prop.iterchildren())
        assert len(bool_prop_values) == 1
        prop_val = bool_prop_values[0]
        assert prop_val.tag == f"{DASCH_SCHEMA}boolean"
        assert prop_val.text == "true"
        assert prop_val.attrib == {"comment": "cmnt", "permissions": "open"}

    def test_region(self):
        xml_root = XMLRoot.create_new("0000", "test")
        xml_root.add_resource(
            RegionResource.create_new(
                "region",
                "lbl",
                "region_of_val",
                Permissions.RESTRICTED,
            ).add_rectangle((0.1, 0.1), (0.2, 0.2))
        )
        serialised = xml_root.serialise(default_permissions=Permissions.OPEN)
        found = list(serialised.iterdescendants(tag=f"{DASCH_SCHEMA}region"))
        assert len(found) == 1
        resource = found[0]
        res_attribs = {"label": "lbl", "id": "region", "permissions": "restricted"}
        assert resource.attrib == res_attribs
        assert len(resource) == 3
        # Check color
        all_colors = list(resource.iterchildren(tag=f"{DASCH_SCHEMA}color-prop"))
        assert len(all_colors) == 1
        color_prop = all_colors[0]
        assert color_prop.attrib == {"name": "hasColor"}
        color_child = next(color_prop.iterchildren())
        assert color_child.text == "#5b24bf"
        assert color_child.attrib == {"permissions": "restricted"}
        # Check is region of
        all_resptrs = list(resource.iterchildren(tag=f"{DASCH_SCHEMA}resptr-prop"))
        assert len(all_resptrs) == 1
        resptr_prop = all_resptrs[0]
        assert resptr_prop.attrib == {"name": "isRegionOf"}
        resptr_child = next(resptr_prop.iterchildren())
        assert resptr_child.text == "region_of_val"
        assert resptr_child.attrib == {"permissions": "restricted"}
        # Check geometry
        all_geo = list(resource.iterchildren(tag=f"{DASCH_SCHEMA}geometry-prop"))
        assert len(all_geo) == 1
        geo_prop = all_geo[0]
        assert geo_prop.attrib == {"name": "hasGeometry"}
        geo_child = next(geo_prop.iterchildren())
        assert geo_child.text
        assert len(geo_child.text) > 0
        assert geo_child.attrib == {"permissions": "restricted"}

    def test_link_resource(self):
        xml_root = XMLRoot.create_new("0000", "test")
        xml_root.add_resource(
            LinkResource.create_new(
                "link_id",
                "lbl",
                ["link1", "link2"],
            ).add_comment("cmnt", Permissions.RESTRICTED)
        )
        serialised = xml_root.serialise(default_permissions=Permissions.OPEN)
        found = list(serialised.iterdescendants(tag=f"{DASCH_SCHEMA}link"))
        assert len(found) == 1
        resource = found[0]
        res_attribs = {"label": "lbl", "id": "link_id", "permissions": "open"}
        assert resource.attrib == res_attribs
        vals = list(resource.iterchildren())
        assert len(vals) == 2
        # link tos
        all_resptrs = list(resource.iterchildren(tag=f"{DASCH_SCHEMA}resptr-prop"))
        assert len(all_resptrs) == 1
        resptr = all_resptrs[0]
        assert resptr.attrib == {"name": "hasLinkTo"}
        for v in resptr.iterchildren():
            assert v.tag == f"{DASCH_SCHEMA}resptr"
            assert v.text in ["link1", "link2"]
            assert v.attrib == {"permissions": "open"}
        # comment
        all_text = list(resource.iterchildren(tag=f"{DASCH_SCHEMA}text-prop"))
        assert len(all_text) == 1
        text_prop = all_text[0]
        assert text_prop.attrib == {"name": "hasComment"}
        text_child = next(text_prop.iterchildren())
        assert text_child.text == "cmnt"
        assert text_child.attrib == {"permissions": "restricted", "encoding": "xml"}

    def test_audio_segment(self):
        xml_root = XMLRoot.create_new("0000", "test")
        xml_root.add_resource(
            AudioSegmentResource.create_new("audio_id", "lbl", "segment_of", 1, 2)
            .add_title("title")
            .add_comment("cmnt")
            .add_description("desc")
            .add_keyword("keywrd")
            .add_relates_to("relates_to")
        )
        serialised = xml_root.serialise(default_permissions=Permissions.OPEN)
        found = list(serialised.iterdescendants(tag=f"{DASCH_SCHEMA}audio-segment"))
        assert len(found) == 1
        resource = found[0]
        res_attribs = {"label": "lbl", "id": "audio_id", "permissions": "open"}
        assert resource.attrib == res_attribs
        assert len(resource) == 7
        segment_of = next(resource.iter(tag=f"{DASCH_SCHEMA}isSegmentOf"))
        assert segment_of.attrib == {"permissions": "open"}
        assert segment_of.text == "segment_of"
        bounds = next(resource.iter(tag=f"{DASCH_SCHEMA}hasSegmentBounds"))
        assert not bounds.text
        assert bounds.attrib == {"segment_start": "1", "segment_end": "2", "permissions": "open"}
        title = next(resource.iter(tag=f"{DASCH_SCHEMA}hasTitle"))
        assert title.attrib == {"permissions": "open"}
        assert title.text == "title"
        comment = next(resource.iter(tag=f"{DASCH_SCHEMA}hasComment"))
        assert comment.attrib == {"permissions": "open"}
        assert comment.text == "cmnt"
        description = next(resource.iter(tag=f"{DASCH_SCHEMA}hasDescription"))
        assert description.attrib == {"permissions": "open"}
        assert description.text == "desc"
        keywrd = next(resource.iter(tag=f"{DASCH_SCHEMA}hasKeyword"))
        assert keywrd.attrib == {"permissions": "open"}
        assert keywrd.text == "keywrd"
        relates_to = next(resource.iter(tag=f"{DASCH_SCHEMA}relatesTo"))
        assert relates_to.attrib == {"permissions": "open"}
        assert relates_to.text == "relates_to"

    def test_video_segment_with_permissions(self):
        xml_root = XMLRoot.create_new("0000", "test")
        xml_root.add_resource(
            VideoSegmentResource.create_new(
                "video_id",
                "lbl",
                "segment_of",
                1,
                2,
                permissions=Permissions.RESTRICTED,
            )
        )
        serialised = xml_root.serialise(default_permissions=Permissions.OPEN)
        found = list(serialised.iterdescendants(tag=f"{DASCH_SCHEMA}video-segment"))
        assert len(found) == 1
        resource = found[0]
        res_attribs = {"label": "lbl", "id": "video_id", "permissions": "restricted"}
        assert resource.attrib == res_attribs
        assert len(resource) == 2
        segment_of = next(resource.iter(tag=f"{DASCH_SCHEMA}isSegmentOf"))
        assert segment_of.attrib == {"permissions": "restricted"}
        assert segment_of.text == "segment_of"
        bounds = next(resource.iter(tag=f"{DASCH_SCHEMA}hasSegmentBounds"))
        assert not bounds.text
        assert bounds.attrib == {"segment_start": "1", "segment_end": "2", "permissions": "restricted"}


def test_root_add_resources() -> None:
    xml_root = XMLRoot.create_new("0000", "test")

    # add_resource
    region_res = RegionResource.create_new("regionID", "label", "id1")
    region_res.add_rectangle((0.1, 0.1), (0.2, 0.2))
    xml_root.add_resource(region_res)
    assert len(xml_root.resources) == 1

    # add_resource_multiple
    many_resources = [Resource.create_new("id1", ":Restype", "label"), Resource.create_new("id2", ":Restype", "label")]
    xml_root.add_resource_multiple(many_resources)
    assert len(xml_root.resources) == 3

    xml_root.add_resource_multiple([])
    assert len(xml_root.resources) == 3

    # add_resource_optional
    no_resource = None
    xml_root.add_resource_optional(no_resource)
    assert len(xml_root.resources) == 3

    serialised = xml_root.serialise()
    region = list(serialised.iterdescendants(tag=f"{DASCH_SCHEMA}region"))
    assert len(region) == 1
    general_resources = list(serialised.iterdescendants(tag=f"{DASCH_SCHEMA}resource"))
    assert len(general_resources) == 2


def test_make_authorship_lookup() -> None:
    res1 = Resource.create_new("id1", ":Restype", "label").add_file(
        "file.jpg", LicenseRecommended.DSP.UNKNOWN, "copy", ["auth", "auth1"]
    )
    res2 = Resource.create_new("id2", ":Restype", "label").add_file(
        "file.jpg", LicenseRecommended.DSP.UNKNOWN, "copy", ["auth2"]
    )
    res3 = Resource.create_new("id2", ":Restype", "label").add_file(
        "file.jpg", LicenseRecommended.DSP.UNKNOWN, "copy", ["auth2"]
    )
    res_none = Resource.create_new("id2", ":Restype", "label").add_file(
        "file.jpg",
    )
    region_res = RegionResource.create_new("regionID", "label", "id1")
    result = _make_authorship_lookup([res1, res2, res3, region_res, res_none])
    assert set(result.lookup.keys()) == {("auth", "auth1"), tuple(["auth2"])}


def test_serialise_authorship() -> None:
    lookup = {("auth", "auth1"): "authorship_1", tuple(["auth2"]): "authorship_2"}
    expected = [
        b'<authorship xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'id="authorship_1"><author>auth</author><author>auth1</author></authorship>',
        b'<authorship xmlns="https://dasch.swiss/schema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        b'id="authorship_2"><author>auth2</author></authorship>',
    ]
    serialised = _serialise_authorship(lookup)
    result = sorted([etree.tostring(x) for x in serialised])
    assert len(result) == len(expected)
    for res, ex in zip(result, expected):
        assert res == ex



"""
This module implements the handling (CRUD) of DSP projects.

CREATE:
    * Instantiate a new object of the class Project with all required parameters
    * Call the ``create``-method on the instance

READ:
    * Instantiate a new object with ``iri`` given
    * Call the ``read``-method on the instance
    * Access the information that has been provided to the instance

"""

from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Union
from urllib.parse import quote_plus

from dsp_tools.clients.connection import Connection
from dsp_tools.commands.project.legacy_models.model import Model
from dsp_tools.error.exceptions import BaseError
from dsp_tools.legacy_models.langstring import LangString


class Project(Model):
    """
    This class represents a project in DSP.

    Attributes
    ----------

    con : Connection
        A Connection instance to a DSP server

    iri : str
        IRI of the project [readonly, cannot be modified after creation of instance]

    shortcode : str
        DSP project shortcode [readonly, cannot be modified after creation of instance]

    shortname : str
        DSP project shortname [read/write]

    longname : str
        DSP project longname [read/write]

    description : LangString
        DSP project description in a given language (Languages.EN, Languages.DE, Languages.FR, Languages.IT).

    keywords : set[str]
        Set of keywords describing the project.

    ontologies : set[str]
        Set if IRI's of the ontologies attached to the project [readonly]

    selfjoin : bool
        Boolean if the project allows selfjoin


    Methods
    -------

    create : DSP project information object
        Creates a new project and returns the information from the project as it is in DSP

    read : DSP project information object
        Read project data from an existing project

    getAllprojects [static]: List of all projects
        Returns a list of all projects available

    print : None
        Prints the project information to stdout

    """

    ROUTE: str = "/admin/projects"
    IRI: str = ROUTE + "/iri/"

    _iri: str
    _shortcode: str
    _shortname: str
    _longname: str
    _description: LangString
    _keywords: set[str]
    _ontologies: set[str]
    _enabled_licenses: set[str]
    _selfjoin: bool
    _status: bool
    _logo: Optional[str]

    SYSTEM_PROJECT: str = "http://www.knora.org/ontology/knora-admin#SystemProject"

    def __init__(
        self,
        con: Connection,
        iri: Optional[str] = None,
        shortcode: Optional[str] = None,
        shortname: Optional[str] = None,
        longname: Optional[str] = None,
        description: LangString = None,
        keywords: Optional[set[str]] = None,
        ontologies: Optional[set[str]] = None,
        enabled_licenses: Optional[set[str]] = None,
        selfjoin: Optional[bool] = None,
        status: Optional[bool] = None,
        logo: Optional[str] = None,
    ):
        """
        Constructor for Project

        :param con: Connection instance
        :param iri: IRI of the project [required for CREATE, READ]
        :param shortcode: Shortcode of the project. Four-digit hexadecimal number. [required for CREATE]
        :param shortname: Shortname of the project [required for CREATE]
        :param longname: Longname of the project [required for CREATE]
        :param description: LangString instance containing the description [required for CREATE]
        :param keywords: Set of keywords [required for CREATE]
        :param ontologies: Set of ontologies that belong to this project [optional]
        :param enabled_licenses: Set of enabled licenses [optional]
        :param selfjoin: Allow selfjoin [required for CREATE]
        :param status: Status of project (active if True) [required for CREATE]
        :param logo: Path to logo image file [optional] NOT YET USED
        """
        super().__init__(con)
        self._iri = iri
        self._shortcode = shortcode
        self._shortname = shortname
        self._longname = longname
        self._description = LangString(description)
        self._keywords = keywords
        if not isinstance(ontologies, set) and ontologies is not None:
            raise BaseError("Ontologies must be a set of strings or None!")
        self._ontologies = ontologies
        self._enabled_licenses = enabled_licenses or set()
        self._selfjoin = selfjoin
        self._status = status
        self._logo = logo

    def __str__(self) -> str:
        tmpstr = self._iri + "\n  " + self._shortcode + "\n  " + self._shortname
        return tmpstr

    #
    # Here follows a list of getters/setters
    #
    @property
    def iri(self) -> Optional[str]:
        return self._iri

    @property
    def shortcode(self) -> Optional[str]:
        return self._shortcode

    @property
    def shortname(self) -> Optional[str]:
        return self._shortname

    @shortname.setter
    def shortname(self, value: str) -> None:
        if self._shortname != str(value):
            self._shortname = str(value)
            self._changed.add("shortname")

    @property
    def longname(self) -> Optional[str]:
        return self._longname

    @longname.setter
    def longname(self, value: str) -> None:
        if self._longname != str(value):
            self._longname = str(value)
            self._changed.add("longname")

    @property
    def description(self) -> LangString:
        return self._description or LangString({})

    @description.setter
    def description(self, value: Optional[LangString]) -> None:
        self._description = LangString(value)
        self._changed.add("description")

    @property
    def keywords(self) -> set[str]:
        return self._keywords

    @keywords.setter
    def keywords(self, value: Union[list[str], set[str]]) -> None:
        if isinstance(value, set):
            self._keywords = value
            self._changed.add("keywords")
        elif isinstance(value, list):
            self._keywords = set(value)
            self._changed.add("keywords")
        else:
            raise BaseError("Must be a set of strings!")

    @property
    def ontologies(self) -> set[str]:
        return self._ontologies

    @property
    def selfjoin(self) -> Optional[bool]:
        return self._selfjoin

    @selfjoin.setter
    def selfjoin(self, value: bool) -> None:
        if self._selfjoin != value:
            self._changed.add("selfjoin")
            self._selfjoin = value

    @property
    def status(self) -> bool:
        return self._status

    @status.setter
    def status(self, value: bool) -> None:
        if self._status != value:
            self._status = value
            self._changed.add("status")

    @property
    def logo(self) -> str:
        return self._logo

    @logo.setter
    def logo(self, value: str) -> None:
        if self._logo != value:
            self._logo = value
            self._changed.add("logo")

    @classmethod
    def fromJsonObj(cls, con: Connection, json_obj: Any) -> Project:
        """
        Internal method! Should not be used directly!

        This method is used to create a Project instance from the JSON data returned by DSP

        :param con: Connection instance
        :param json_obj: JSON data returned by DSP as python3 object
        :return: Project instance
        """
        iri = json_obj.get("id")
        if iri is None:
            raise BaseError("Project iri is missing")
        shortcode = json_obj.get("shortcode")
        if shortcode is None:
            raise BaseError("Shortcode is missing")
        shortname = json_obj.get("shortname")
        if shortname is None:
            raise BaseError("Shortname is missing")
        longname = json_obj.get("longname")
        if longname is None:
            raise BaseError("Longname is missing")
        description = LangString.fromJsonObj(json_obj.get("description"))
        keywords = set(json_obj.get("keywords"))
        if keywords is None:
            raise BaseError("Keywords are missing")
        ontologies = set(json_obj.get("ontologies"))
        if ontologies is None:
            raise BaseError("ontologies are missing")
        enabled_licenses = json_obj.get("enabledLicenses", set())
        selfjoin = json_obj.get("selfjoin")
        if selfjoin is None:
            raise BaseError("Selfjoin is missing")
        status = json_obj.get("status")
        if status is None:
            raise BaseError("Status is missing")
        logo = json_obj.get("logo")
        return cls(
            con=con,
            iri=iri,
            shortcode=shortcode,
            shortname=shortname,
            longname=longname,
            description=description,
            keywords=keywords,
            ontologies=ontologies,
            enabled_licenses=enabled_licenses,
            selfjoin=selfjoin,
            status=status,
            logo=logo,
        )

    def createDefinitionFileObj(self) -> dict[str, Any]:
        return {
            "shortcode": self._shortcode,
            "shortname": self._shortname,
            "longname": self._longname,
            "descriptions": self._description.createDefinitionFileObj(),
            "keywords": list(self._keywords),
            "enabled_licenses": list(self._enabled_licenses),
        }

    def create(self) -> Project:
        """
        Create a new project in DSP

        :return: JSON-object from DSP
        """
        jsonobj = self._toJsonObj_create()
        result = self._con.post(Project.ROUTE, jsonobj)
        return Project.fromJsonObj(self._con, result["project"])

    def _toJsonObj_create(self) -> dict[str, str]:
        tmp = {}
        if self._shortcode is None:
            raise BaseError("There must be a valid project shortcode!")
        tmp["shortcode"] = self._shortcode
        if self._shortname is None:
            raise BaseError("There must be a valid project shortname!")
        tmp["shortname"] = self._shortname
        if self._longname is None:
            raise BaseError("There must be a valid project longname!")
        tmp["longname"] = self._longname
        if self._description.isEmpty():
            raise BaseError("There must be a valid project description!")
        tmp["description"] = self._description.toJsonObj()
        if self._keywords is not None and len(self._keywords) > 0:
            tmp["keywords"] = self._keywords
        if self._enabled_licenses:
            tmp["enabledLicenses"] = list(self._enabled_licenses)
        if self._selfjoin is None:
            raise BaseError("selfjoin must be defined (True or False!")
        tmp["selfjoin"] = self._selfjoin
        if self._status is None:
            raise BaseError("status must be defined (True or False!")
        tmp["status"] = self._status
        return tmp

    def read(self) -> Project:
        """
        Read a project from DSP

        :return: JSON-object from DSP
        """
        result = None
        if self._iri is not None:
            result = self._con.get(Project.IRI + quote_plus(self._iri))
        elif self._shortcode is not None:
            result = self._con.get(Project.ROUTE + "/shortcode/" + quote_plus(self._shortcode))
        elif self._shortname is not None:
            result = self._con.get(Project.ROUTE + "/shortname/" + quote_plus(self._shortname))
        if result is not None:
            return Project.fromJsonObj(self._con, result["project"])
        else:
            raise BaseError(
                f"ERROR: Could not read project '{self.shortname}' ({self.shortcode}) with IRI {self._iri} "
                f"from DSP server."
            )

    @staticmethod
    def getAllProjects(con: Connection) -> list[Project]:
        """
        Get all existing projects in DSP

        :param con: Connection instance
        :return:
        """
        try:
            result = con.get(Project.ROUTE)
            return [Project.fromJsonObj(con, a) for a in result["projects"]]
        except BaseError:
            return []



# mypy: disable-error-code="no-untyped-def"

from pathlib import Path

import pytest

from dsp_tools.cli.args import ServerCredentials
from dsp_tools.cli.args import ValidateDataConfig
from dsp_tools.cli.args import ValidationSeverity
from dsp_tools.clients.authentication_client import AuthenticationClient
from dsp_tools.clients.authentication_client_live import AuthenticationClientLive
from dsp_tools.commands.project.create.project_create_all import create_project
from dsp_tools.commands.validate_data.api_clients import ShaclValidator
from dsp_tools.commands.validate_data.get_user_validation_message import sort_user_problems
from dsp_tools.commands.validate_data.models.input_problems import OntologyValidationProblem
from dsp_tools.commands.validate_data.models.input_problems import ProblemType
from dsp_tools.commands.validate_data.models.validation import ValidationReportGraphs
from dsp_tools.commands.validate_data.query_validation_result import reformat_validation_graph
from dsp_tools.commands.validate_data.validate_data import _get_validation_result
from dsp_tools.commands.validate_data.validate_data import _prepare_data_for_validation_from_file
from dsp_tools.commands.validate_data.validate_ontology import validate_ontology

# ruff: noqa: ARG001 Unused function argument

CONFIG = ValidateDataConfig(Path(), None, ValidationSeverity.INFO, False)


@pytest.fixture(scope="module")
def _create_projects_edge_cases(creds: ServerCredentials) -> None:
    assert create_project(Path("testdata/validate-data/special_characters/project_special_characters.json"), creds)
    assert create_project(Path("testdata/validate-data/inheritance/project_inheritance.json"), creds)
    assert create_project(Path("testdata/validate-data/erroneous_ontology/project_erroneous_ontology.json"), creds)


@pytest.fixture(scope="module")
def authentication(creds: ServerCredentials) -> AuthenticationClient:
    auth = AuthenticationClientLive(server=creds.server, email=creds.user, password=creds.password)
    return auth


@pytest.fixture(scope="module")
def special_characters_violation(
    _create_projects_edge_cases, authentication: AuthenticationClient, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/special_characters/special_characters_violation.xml")
    graphs, _ = _prepare_data_for_validation_from_file(file, authentication)
    return _get_validation_result(graphs, shacl_validator, CONFIG)


@pytest.fixture(scope="module")
def inheritance_violation(
    _create_projects_edge_cases, authentication: AuthenticationClient, shacl_validator: ShaclValidator
) -> ValidationReportGraphs:
    file = Path("testdata/validate-data/inheritance/inheritance_violation.xml")
    graphs, _ = _prepare_data_for_validation_from_file(file, authentication)
    return _get_validation_result(graphs, shacl_validator, CONFIG)


@pytest.fixture(scope="module")
def validate_ontology_violation(
    _create_projects_edge_cases, authentication: AuthenticationClient, shacl_validator: ShaclValidator
) -> OntologyValidationProblem | None:
    file = Path("testdata/validate-data/erroneous_ontology/erroneous_ontology.xml")
    graphs, _ = _prepare_data_for_validation_from_file(file, authentication)
    return validate_ontology(graphs.ontos, shacl_validator, CONFIG)


@pytest.mark.usefixtures("_create_projects_edge_cases")
def test_special_characters_correct(authentication: AuthenticationClient, shacl_validator: ShaclValidator) -> None:
    file = Path("testdata/validate-data/special_characters/special_characters_correct.xml")
    graphs, _ = _prepare_data_for_validation_from_file(file, authentication)
    special_characters_correct = _get_validation_result(graphs, shacl_validator, CONFIG)
    assert special_characters_correct.conforms


def test_special_characters_violation(special_characters_violation: ValidationReportGraphs) -> None:
    assert not special_characters_violation.conforms


def test_reformat_special_characters_violation(special_characters_violation: ValidationReportGraphs) -> None:
    expected_tuples = [
        (
            "node_backslash",
            (
                "A valid node from the list 'list \\ ' space' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
            "list \\ ' space / other \\ backslash",
        ),
        (
            "node_double_quote",
            (
                "A valid node from the list 'list \\ ' space' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
            '''list \\ ' space / other double quote "''',
        ),
        (
            "node_single_quote",
            (
                "A valid node from the list 'list \\ ' space' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
            "list \\ ' space / other single quote '",
        ),
        ("non_ascii_latin_alphabet", "", ""),
        ("non_ascii_other_alphabet", "", ""),
        ("special_char", "", ""),
        (
            "wrong_list_name",
            (
                "A valid node from the list 'list \\ ' space' must be used with this property "
                "(input displayed in format 'listName / NodeName')."
            ),
            "other / \\ backslash",
        ),
    ]
    result = reformat_validation_graph(special_characters_violation)
    sorted_problems = sort_user_problems(result)
    assert len(sorted_problems.unique_violations) == len(expected_tuples)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(result.problems, key=lambda x: x.res_id)
    for prblm, expected in zip(alphabetically_sorted, expected_tuples):
        if prblm.problem_type == ProblemType.GENERIC:
            assert prblm.res_id == expected[0]
            assert prblm.message == expected[1]
            assert prblm.input_value == expected[2]
        elif prblm.problem_type == ProblemType.INPUT_REGEX:
            assert prblm.res_id == expected[0]


@pytest.mark.usefixtures("_create_projects_edge_cases")
def test_inheritance_correct(authentication: AuthenticationClient, shacl_validator: ShaclValidator) -> None:
    file = Path("testdata/validate-data/inheritance/inheritance_correct.xml")
    graphs, _ = _prepare_data_for_validation_from_file(file, authentication)
    inheritance_correct = _get_validation_result(graphs, shacl_validator, CONFIG)
    assert inheritance_correct.conforms


def test_inheritance_violation(inheritance_violation: ValidationReportGraphs) -> None:
    assert not inheritance_violation.conforms


def test_reformat_inheritance_violation(inheritance_violation: ValidationReportGraphs) -> None:
    expected_results = [
        ("ResourceSubCls1", {"onto:hasText0"}),
        ("ResourceSubCls2", {"onto:hasTextSubProp1", "onto:hasText0"}),
        ("ResourceSubCls2", {"onto:hasTextSubProp1", "onto:hasText0"}),
        ("ResourceUnrelated", {"onto:hasText0"}),
    ]
    result = reformat_validation_graph(inheritance_violation)
    sorted_problems = sort_user_problems(result)
    assert len(sorted_problems.unique_violations) == len(expected_results)
    assert not sorted_problems.user_warnings
    assert not sorted_problems.user_info
    assert not sorted_problems.unexpected_shacl_validation_components
    alphabetically_sorted = sorted(result.problems, key=lambda x: x.res_id)
    for one_result, expected in zip(alphabetically_sorted, expected_results):
        assert one_result.problem_type == ProblemType.NON_EXISTING_CARD
        assert one_result.res_id == expected[0]
        assert one_result.prop_name in expected[1]


def test_validate_ontology_violation(validate_ontology_violation: ValidationReportGraphs | None) -> None:
    assert isinstance(validate_ontology_violation, OntologyValidationProblem)
    erroneous_cards_msg = {
        "isPartOf must either have cardinality 1 or 0-1.",
        "seqnum must either have cardinality 1 or 0-1.",
    }
    missing_is_part_of = {"A class with a cardinality for seqnum also requires a cardinality for isPartOf."}
    missing_seqnum = {"A class with a cardinality for isPartOf also requires a cardinality for seqnum."}
    mixed_cards = {"The cardinalities for seqnum and isPartOf must be identical within one resource class."}
    expected_results = [
        ("error:ImageWithKnoraProp_ErroneousCards", erroneous_cards_msg),
        ("error:ImageWithKnoraProp_ErroneousCards", erroneous_cards_msg),
        ("error:ImageWithKnoraProp_MissingIsPartOf", missing_is_part_of),
        ("error:ImageWithKnoraProp_MissingSeqnum", missing_seqnum),
        ("error:ImageWithKnoraProp_MixedValidCards", mixed_cards),
        ("error:ImageWithSubProp_ErroneousCards", erroneous_cards_msg),
        ("error:ImageWithSubProp_ErroneousCards", erroneous_cards_msg),
        ("error:ImageWithSubProp_MissingIsPartOf", missing_is_part_of),
        ("error:ImageWithSubProp_MissingSeqnum", missing_seqnum),
        ("error:ImageWithSubProp_MixedValidCards", mixed_cards),
    ]
    sorted_problems = sorted(validate_ontology_violation.problems, key=lambda x: x.res_iri)
    assert len(validate_ontology_violation.problems) == len(expected_results)
    for one_result, expected in zip(sorted_problems, expected_results):
        assert one_result.res_iri == expected[0]
        assert one_result.msg in expected[1]



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import cast

import regex

from dsp_tools.error.exceptions import BaseError

_calendar = r"GREGORIAN|JULIAN|ISLAMIC"
_era = r"CE|BCE|BC|AD"
_year = r"\d{1,4}"
_month = r"\d{1,2}"
_day = r"\d{1,2}"
_full_date_pattern = rf"""
^
(?:({_calendar}):)?                         # optional calendar
(?:({_era}):)?                              # optional era
({_year}(?:-{_month})?(?:-{_day})?)         # date
(?::({_era}))?                              # optional era
(?::({_year}(?:-{_month})?(?:-{_day})?))?   # optional date
$
"""
_single_date_pattern = rf"^({_year})(?:-({_month}))?(?:-({_day}))?$"


class Calendar(Enum):
    """Enum for calendar types."""

    GREGORIAN = "GREGORIAN"
    ISLAMIC = "ISLAMIC"
    JULIAN = "JULIAN"

    @staticmethod
    def from_string(s: str) -> Calendar:
        """Parses a string into a calendar type, potentially failing with a BaseError."""
        match s:
            case "GREGORIAN":
                return Calendar.GREGORIAN
            case "ISLAMIC":
                return Calendar.ISLAMIC
            case "JULIAN":
                return Calendar.JULIAN
            case _:
                raise BaseError(f"Invalid calendar type: {s}")


class Era(Enum):
    """Enum for era types."""

    AD = "AD"
    BC = "BC"
    BCE = "BCE"
    CE = "CE"

    @staticmethod
    def from_string(s: str) -> Era:
        """Parses a string into an era type, potentially failing with a BaseError."""
        match s:
            case "AD":
                return Era.AD
            case "BC":
                return Era.BC
            case "BCE":
                return Era.BCE
            case "CE":
                return Era.CE
            case _:
                raise BaseError(f"Invalid era type: {s}")


class DayMonthYearEra(Enum):
    DAY = "Day"
    MONTH = "Month"
    YEAR = "Year"
    ERA = "Era"


class StartEnd(Enum):
    START = "Start"
    END = "End"


@dataclass(frozen=True)
class SingleDate:
    """Information about a single date."""

    era: Era | None
    year: int
    month: int | None
    day: int | None


@dataclass(frozen=True)
class Date:
    """Information about a date."""

    calendar: Calendar
    start: SingleDate
    end: SingleDate | None


def parse_date_string(s: str) -> Date:
    """Parse a date string into a Date object.

    Args:
        s: date string

    Returns:
        Date object

    Raises:
        BaseError: if the date string cannot be parsed
    """

    calendar, start_era, start_date, end_era, end_date = _split_date_string(s)
    calendar_enum = Calendar.from_string(calendar or "GREGORIAN")
    if not end_date:
        end_date = start_date
    if calendar_enum != Calendar.ISLAMIC:
        if not start_era:
            start_era = "CE"
        if end_date and not end_era:
            end_era = start_era
    start_era_enum = Era.from_string(start_era) if start_era else None
    end_era_enum = Era.from_string(end_era) if end_era else None
    start = _parse_single_date(start_date, start_era_enum)
    end = _parse_single_date(end_date, end_era_enum) if end_date else None

    return Date(calendar_enum, start, end)


def is_full_date(s: str) -> bool:
    """
    Check if a string is a full DSP date string of the scheme calendar:epoch:yyyy-mm-dd:epoch:yyyy-mm-dd.
    """
    return bool(regex.search(_full_date_pattern, s, flags=regex.VERBOSE))


def _split_date_string(s: str) -> tuple[str | None, str | None, str, str | None, str | None]:
    date_match = regex.search(_full_date_pattern, s, flags=regex.VERBOSE)
    if not date_match:
        raise BaseError(f"Could not parse date: {s}")
    date_groups = date_match.groups()
    match date_groups:
        case ("ISLAMIC", start_era, _, end_era, _) if start_era or end_era:
            raise BaseError(f"ISLAMIC calendar does not support eras: {s}")
        case (str() | None, str() | None, str(), str() | None, str() | None):
            return cast(tuple[str | None, str | None, str, str | None, str | None], date_groups)
        case _:
            raise BaseError(f"Could not parse date: {s}")


def _parse_single_date(date: str, era: Era | None) -> SingleDate:
    date_match = regex.search(_single_date_pattern, date)
    if not date_match:
        raise BaseError(f"Could not parse date: {date}")
    match date_match.groups():
        case (str() as year, str() | None as month, str() | None as day):
            y = int(year)
            m = int(month) if month else None
            d = int(day) if day else None
            return SingleDate(era, y, m, d)
        case _:
            raise BaseError(f"Could not parse date: {date}")
